"""Grid associated to site-based data.

This Grid type actually represents the absence of a grid. It is meant to describe a
collection of unrelated sites or locations. Spatial reduction operations are only defined
for both axes (axis = 'both').
"""

import numpy as np
import pandas as pd
import copy
from typing import Optional, Sequence, Tuple, SupportsFloat
from canopy.grid.grid_abc import Grid
from canopy.grid.grid_empty import GridEmpty
from canopy.grid.registry import register_grid, register_gridop, get_gridop
import random

grid_type = 'sites'

def _setindex(grid: 'GridSites', new_index: Sequence):
    grid.sites.index = new_index
    grid.sites.index.name = "Key"
    grid.sites.sort_index(inplace=True)

@register_grid(grid_type)
class GridSites(Grid):
    """A subclass of Grid used to represent a collection of sites.

    Parameters
    ----------
    sites
        The site locations are arranged in rows. For example:
        .. code-block::
        
            [[12.25, -4.75],
             [-27.75, 45.25],
             [-50.25, 73.25]]
    axis0 : str
        Name of first axis (default: 'lon').
    axis1 : str
        Name of second axis (default: 'lat').
    """

    def __init__(self, sites: np.ndarray | None, names: Sequence[str | None] | None = None, axis0: str = 'lon', axis1: str = 'lat', gridop: str | None = None) -> None:
        super().__init__(grid_type, axis0=axis0, axis1=axis1)

        if gridop is None and sites is None:
            raise ValueError("'sites' and 'gridop' cannot be both None")

        if sites is None:
            self.sites = None
            get_gridop(self.grid_type, gridop, 'both') # KeyError if gridop doesn't exist
            self.axis_gridop['both'] = gridop
            self.axis_gridop[self.axis_names[0]] = gridop
            self.axis_gridop[self.axis_names[1]] = gridop
        else:
            self.axis_gridop['both'] = None
            if not isinstance(sites, np.ndarray):
                raise ValueError("'sites' must be a numpy array")
            if sites.shape[1] != 2:
                raise ValueError("'sites' must have exactly two columns, corresponding to the spatial axes")
            if len(sites) < 1:
                raise ValueError("'sites' must have at least one row")
            sites = sites.astype(float) # Will fail if not numbers
            index = []
            if names is None:
                for site in sites:
                    index.append(tuple(site))
            else:
                if len(sites) != len(names):
                    raise ValueError("The number of rows of the 'sites' array and the length of the 'names' list must be equal.")
                for name, site in zip(names, sites):
                    index.append(str(name) if not name is None else tuple(site))

            self.sites = pd.DataFrame(sites, dtype=float, columns=[axis0, axis1])
            _setindex(self, index)


    @classmethod
    def from_frame(cls, df):
        """Create a GridSites instance from a DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            A pandas DataFrame with a valid format (see Field documentation).

        Returns
        -------
        An instance of the grid subclass.
        """
        index = df.index
        axis0, axis1 = index.names[:2]
        sites = np.array(list(index.droplevel('time').drop_duplicates().values))
        return cls(axis0 = axis0, axis1 = axis1, sites=sites)


    def to_txt(self, fname: str, sep: str = ' ', randomize: bool = False) -> None:
        lines = []
        for key, coords in zip(self.sites.index, self.sites.values):
            coord0 = str(coords[0])
            coord1 = str(coords[1])
            line = coord0 + sep + coord1
            if isinstance(key, str):
                line += sep + key
            line += '\n'
            lines.append(line)

        if randomize:
            random.shuffle(lines)

        with open(fname, 'w') as f:
            f.writelines(lines)


    def get_reduced_grid(self, gridop: str, axis: str) -> Grid:
        """Create a new grid, reduced according to the parameters

        Parameters
        ----------
        gridop : str
            The reduction operation
        axis : str
            The axis to be reduced

        Returns
        -------
        An instance of GridSites
        """
        get_gridop(self.grid_type, gridop, axis) # Will raise a KeyError if it doesn't exist
        return GridSites(sites = None, gridop = gridop, axis0 = self.axis_names[0], axis1 = self.axis_names[1])


    def get_sliced_grid(self,
                        axis0_slice: Optional[tuple[float,float]] = None,
                        axis1_slice: Optional[tuple[float,float]] = None) -> Grid:
        """Create a new grid, sliced according to the parameters.

        Parameters
        ----------
        axis0_slice : tuple[float,float]
            Specifies an interval on axis0.
        axis1_slice : tuple[float,float]
            Specifies an interval on axis1.

        Returns
        -------
        An instance of the grid subclass.
        """
        sites = self.sites
        if axis0_slice is not None:
            x0min, x0max = axis0_slice
            ax0 = self.axis_names[0]
            sites = sites[(sites[ax0] >= x0min) & (sites[ax0] <= x0max)]
        if axis1_slice is not None:
            x1min, x1max = axis1_slice
            ax1 = self.axis_names[1]
            sites = sites[(sites[ax1] >= x1min) & (sites[ax1] <= x1max)]

        if sites.empty:
            return GridEmpty()
        else:
            names = [x if isinstance(x,str) else None for x in sites.index]
            return GridSites(sites.values, names = names, axis0 = self.axis_names[0], axis1 = self.axis_names[1])
        

    def is_compatible(self, other) -> bool:

        if self.grid_type != other.grid_type:
            return False

        if self.axis_gridop != other.axis_gridop:
            return False

        return True


    def __add__(self, other):

        if not self.is_compatible(other):
            raise ValueError("Non-compatible grids cannot be aggregated.")

        if self.is_reduced('both'):
            raise ValueError("Reduced grids cannot be added.")

        sites = pd.concat([self.sites, other.sites])
        sites = sites[~sites.index.duplicated(keep='first')]
        names = [x if isinstance(x,str) else None for x in sites.index]

        return GridSites(sites.values, names=names, axis0=self.axis_names[0], axis1=self.axis_names[0])


    def __repr__(self) -> str:
        repr_str = [super().__repr__()]
        if self.is_reduced('both'):
            repr_str.append(f"Grid is reduced (gridop: '{self.axis_gridop['both']}').")
        else:
            repr_str.append(f"Gridlist:\n{str(self.sites)}")
        return '\n'.join(repr_str)


    def __str__(self) -> str:
        return self.__repr__()


    def assign_names(self, names_dict: dict[Tuple[SupportsFloat],str], atol=1.e-6, rtol=0) -> None:
        """Assign names to sites based on their coordinates

        Parameters
        ----------
        names_dict: dict
            A dictionary whose keys are tuples representing the coordinates of a site,
            and whose values (strings) are the names to assign to those sites
        atol: float
            Absolute tolerance for the coordinate comparison, as in numpy's allclose function.
        rtol: float
            Relative tolerance for the coordinate comparison, as in numpy's allclose function.
        """
        named_coords = []
        for k, v in names_dict.items():
            coord1 = np.array(k, dtype=float).flatten()
            if coord1.size != 2:
                raise ValueError(f"Key {k} cannot be interpreted as a coordinate")
            named_coords.append(coord1)

        new_index = []
        for old_key, coord2 in zip(self.sites.index, self.sites.values):
            for coord1, name in zip(named_coords, names_dict.values()):
                found = False
                if np.allclose(coord1, coord2, atol=atol, rtol=rtol):
                    new_index.append(str(name))
                    found = True
                    break
            if not found:
                new_index.append(old_key)

        _setindex(self, new_index)
                    

@register_gridop(grid_type)
def av_both(df: pd.DataFrame, grid: Grid) -> pd.DataFrame:
    """Spatially average the data.
    
    On this 'grid', all sites count the same for the average.

    Parameters
    ----------
    df : pd.DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : GridSites
        A GridSites object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['time']
    df_red = df.groupby(group_levels).mean()
    #df_red = _restore_index(df_red, 'av')

    return df_red


@register_gridop(grid_type)
def sum_both(df: pd.DataFrame, grid: Grid) -> pd.DataFrame:
    """Spatially aggregate the data.

    On this 'grid', all sites count the same for the sum.

    Parameters
    ----------
    df : pd.DataFrame
        The pandas DataFrame whose data is to be aggregated.
    grid : GridSites
        A GridSites object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['time']
    df_red = df.groupby(group_levels).sum()
    #df_red = _restore_index(df_red, 'sum')

    return df_red




# ------------------------------------------
# | Tests for static plotting using canopy |
# ------------------------------------------

import pytest
from pathlib import Path
import matplotlib
matplotlib.use('Agg')  # Set non-GUI backend before pyplot import
from canopy import Field
import canopy.visualization as cv

DATA_DIR = Path("tests/test_data")

@pytest.mark.filterwarnings("ignore:FigureCanvasAgg is non-interactive, and thus cannot be shown:UserWarning")
def test_make_static_plot():
    """Test that make_static_plot works on two test fields."""

    field_a = Field.from_file(DATA_DIR / "anpp_spain_1990_2010.out.gz")
    field_b = Field.from_file(DATA_DIR / "anpp_spain_1990_2010_mod.out.gz")

    cv.make_static_plot(
        field_a=field_a,
        field_b=field_b,
        layers=["Abi_alb","Bet_pen","Bet_pub","Que_rob","C3_gr"],
        field_a_label="no mod",
        field_b_label="mod",
        unit_a="kgC m-2",
        unit_b="kgC m-2",
        title="Actual NPP in Spain (1990-2010)", 
        palette="tab10",
        move_legend = True,
        dark_mode=True, 
        transparent=False,
        x_fig=10,
        y_fig=10
        )


from canopy.visualization.map import make_simple_map, make_diff_map
from canopy.visualization.time_series import make_time_series
from canopy.visualization.static_plot import make_static_plot
from canopy.visualization.comparison_plot import make_comparison_plot
from canopy.visualization.taylor_diagram import make_taylor_diagram
from canopy.visualization.plot_functions import multiple_figs


"""Registry functionality for grids and grid operations.

This module provides:

- A registry of different Grid types: A dictionary where the keys are string identifiers for each
  type of grid (e.g. `'sites'`, `'lonlat'`...) and the values are the corresponding subclass of Grid
  (e.g. `GridSites`, `GridLonLat`...)
- A registry of spatial reduction operations on the different grids: A dictionary where the keys are
  tuples of the form `(grid_type, operation, axis)`, and the values are the functions that perform the
  operation on the dataframe. In the key, `grid_type` is the string identifier of the grid (`'sites'`, `'lonlat'`...),
  `operation` is a string identifying the reduction operation, and `axis` is the axis name as it appears
  on the DataFrame index.

Examples
--------
Suppose we want to register a cartesian grid and its associated operations. The first step is to
create a new file in the grid/ folder. This file will contain the grid description and the supported
grid operations. Let's assume the file is grid/grid_xy.py. The code described below goes in this file.

We first define a new class, GridXY, inheriting from the Grid abstract base class. This class
will be registered with the string identifier passed to the register_grid class decorator:

.. code-block::

    from canopy.grid.grid_abc import Grid
    grid_type = 'xy' # String used to register this grid type.
    @register_grid(grid_type)
    class GridXY(Grid):
        super().__init__(axis0 = 'x', axis1 = 'y')
        ...

The Grid object's documentation details the mandatory abstract methods to implement.

we define the operations that are allowed on this grid, and register them with the
register_gridop decorator. This decorator will use the name of the function to form a key to
identify the operation. The name of the function must be of the format `operation_axis`.
For example, to register an averaging operation along the x axis, we would do:

.. code-block::

    @register_gridop(grid_type)
    def av_x(df: pd.DataFrame, grid: Grid) -> pd.DataFrame:
        ...

Note that grid operations have all the same signature.

Lastly, we need to add the new file, grid/grid_xy.py, to grid/__init.py__:

    import canopy.grid.grid_xy

If all went well, the new grid type and grid operations will be available. For example, the
following should work:

.. code-block::

    from canopy.field import Field
    field = Field.from_file(path, grid_type='xy')
    field.red_space('av', 'x', inplace=True)
    print(field)
"""

import pandas as pd
from typing import Callable, Type
from canopy.grid.grid_abc import Grid

# Grid type registry
_grids: dict[str, Type[Grid]] = {}

# Grid operations registry
_grid_operations: dict[tuple[str, str, str], Callable[[pd.DataFrame, Grid], pd.DataFrame]] = {}


def register_grid(grid_type: str) -> Callable:
    """Add decorated Grid subclass to the grid registry.

    Parameters
    ----------
    grid_type : str
        A string that identifies the type of Grid being registered.

    Returns
    -------
    A decorator to register the class with the provided identifier.
    """
    def decorator(cls: Type[Grid]) -> Type[Grid]:
        if grid_type in _grids:
            raise KeyError(f"Grid type {grid_type} is already registered.")
        else:
            _grids[grid_type] = cls
        return cls

    return decorator


def create_grid(grid_type: str, **kwargs) -> Grid:
    """Create an instance of a Grid object of the specified type.

    Parameters
    ----------
    grid_type : str
        The string identifying the grid type to create.
    **kwargs
        The keyword arguments are forwarded to the selected grid's constructor.

    Returns
    -------
    An instance of a subclass of Grid, specified by the grid_type parameter.
        
    Examples
    --------
    .. code-block::
    
        # Create a 'lonlat' type grid
        from canopy.grid import create_grid
        grid = create_grid('lonlat',
                           lon_min = -12.25, lon_max = 10.75, dlon = 0.5,
                           lat_min = 20.25, lat_max = 40.25, dlat = 0.5)
    """
    try:
        grid = _grids[grid_type](**kwargs)
    except KeyError:
        raise KeyError(f"Grid must be one of {[_ for _ in _grids]}")

    return grid


def get_grid(grid_type: str) -> Type[Grid]:
    """Get a reference to the uninstantiated Grid subclass of the specified type.

    Parameters
    ----------
    grid_type : str
        The grid type identifier (e.g., 'lonlat, 'sites')

    Returns
    -------
    The Grid subclass type registered under the specified grid_type string.
    """
    try:
        grid = _grids[grid_type]
    except KeyError:
        raise KeyError(f"Grid must be one of {[_ for _ in _grids]}")
    return grid


def get_grid_type(grid: Grid) -> str:
    """Get the grid type string identifier of the passed Grid instance.

    Parameters
    ----------
    grid : Grid
        An instance of a Grid subclass.
    
    Returns
    -------
    The grid type's string identifier.
    """
    for k, v in _grids.items():
        if isinstance(grid, v):
            return k
    raise ValueError("Grid type not found in registry.")


def register_gridop(grid_type: str):
    """Add decorated function to the grid operations registry.

    Parameters
    ----------
    grid_type : str
        The grid type string identifier.

    Returns
    -------
    A decorator to register a grid grid operation.

    Notes
    -----
    The decorator forms the registry key (the tuple `(grid_type, operation, axis)`) from the
    argument `grid_type`, supplied to the decorator, and the name of the decorated function,
    which must be of the form `operation_axis`.

    The name of the operation can be any string. But it should make sense and be consistent
    with all the other gridops (for example, if you want to register averaging operations for
    axes `'x'` and `'y'`, don't use `mean_x` and `av_y`).

    The names of the axes, however, must be the consistent with the names of the indices on
    the DataFrame that the operation is meant to act on.
    """
    def decorator(gop):
        op, axis = gop.__name__.split('_')
        key = (grid_type, op, axis)
        if key in _grid_operations:
            raise KeyError(f"Grid operation {key} is already registered.")
        else:
            _grid_operations[key] = gop
        return gop
    return decorator


def get_gridop(grid: str | Grid, gridop: str, axis: str) -> Callable[[pd.DataFrame, Grid], pd.DataFrame]:
    """Retrieve a grid operation function from the registry.
    
    Parameters
    ----------
    grid : str | Grid
        An instance of the Grid object describing the grid type.
    gridop : str
        The string identifying the grid operation (e.g., 'av' for average).
    axis : str
        The name of the axis along which the operation is performed (e.g. 'lon')

    Returns
    -------
    The function that performs the selected grid operation.
    """
    if isinstance(grid, Grid):
        grid_type = get_grid_type(grid)
    elif isinstance(grid, str):
        grid_type = grid
    try:
        return _grid_operations[(grid_type, gridop, axis)]
    except KeyError:
        raise KeyError(f"Operation '{gridop}' on axis '{axis}' is not defined for grid type '{grid_type}'.")




import canopy.grid.grid_sites
import canopy.grid.grid_lonlat
import canopy.grid.grid_empty

from canopy.grid.registry import register_grid, create_grid, get_grid, get_grid_type, register_gridop, get_gridop




import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import warnings
from typing import Optional, List
from canopy.visualization.plot_functions import get_color_palette, make_dark_mode, handle_figure_output, select_sites
import canopy as cp

def make_comparison_plot(fields: cp.Field | List[cp.Field], output_file: Optional[str] = None,
                         plot_type: Optional[str] = "box", layers: Optional[List[str] | str] = None,
                         gridop: Optional[str] = None, sites: Optional[bool | List[tuple]] = False,
                         yaxis_label: Optional[str] = None, field_labels: Optional[List[str]] = None,
                         unit: Optional[List[str]] = None, title: Optional[str] = None, 
                         palette: Optional[str] = None, custom_palette: Optional[List[str]] = None, 
                         horizontal: Optional[bool] = False, vertical_xlabels: Optional[bool] = False, 
                         dark_mode: bool = False, transparent: bool = False, 
                         x_fig: float = 10, y_fig: float = 10, 
                         return_fig: Optional[bool] = False, **kwargs) -> Optional[plt.Figure]:
    """
    Create a comparative plot from a list of input data Fields from, for example, different runs. The functions can generate
    boxplot, strip or swarm plot, violin plot, boxen plot, point plot, bar plot or count plot based on the `plot_type` parameter.

    Parameters
    ----------
    fields : cp.Field or List[cp.Field]
        Input data Field to display.
    output_file : str, optional
        File path for saving the plot.
    plot_type: str, optional
        Type of plot. Either “strip”, “swarm”, “box”, “violin”, “boxen”, “point”, or “bar”
    layers : List[str] or str, optional
        List of layer names to display.
    gridop : str, optional
        If provided, the grid reduction operation. Either None, 'sum' or 'av'. Default is None.
    sites : bool or List[Tuple], optional
        Control site-level plotting instead of spatial reduction. Default is False. True = all sites,
        if provided with a list, only select the sites in the list.
    yaxis_label : str, optional
        Y-axis label, if not provided canopy will try to retrieve the name of the variable in the metadata.
    field_labels : List[str], optional
        Names of each series to display in the legend.
    unit : List[str], optional
        Unit of the y-axis variable, if not provided canopy will try to retrieve the unit of the variable in the metadata.
    title : str, optional
        Title of the plot.
    palette : str, optional
        Seaborn color palette to use for the line colors (https://seaborn.pydata.org/tutorial/color_palettes.html, 
        recommended palette are in https://colorbrewer2.org).
    custom_palette : List[str], optional
        Path of custom color palette .txt file to use. Names should match label names.
    horizontal : bool, optional
        If True, renders the plot with horizontal orientation (flips the axes).
    vertical_xlabels : bool, optional
        If True, rotates the x-axis tick labels vertically (i.e., 90 degrees).
    dark_mode : bool, optional
        If True, applies dark mode styling to the figure. Default is False.
    transparent : bool, optional
        If True, sets the figure background to be transparent when saved. Default is False.
    x_fig : float, optional
        Width of the figure in inches. Default is 10.
    y_fig : float, optional
        Height of the figure in inches. Default is 10.
    return_fig : bool, optional
        If True, returns the figure object that can be usuable by multiple_figs().
        Default is False.
    **kwargs
        Additional keyword arguments are passed directly to `seaborn.catplot`. This allows customization of
        plot features such as `aspect`, `errorbar`, height`, etc.
    """
    # Check for some initial conditions
    if layers and sites:
        raise ValueError("layers and sites argument cannot be used simultanuously. Only one layer for multiple sites.")
    if plot_type == "count":
        raise ValueError("count plot is not supported for comparison plot.")
    if gridop is not None and sites is not False:
        warnings.warn("If gridop is selected, sites won't have an effect because the space is going to be reduced anyway.", UserWarning)
        sites = False

    # Force variables to be a list
    if not isinstance(fields, list):
        fields = [fields]
    if isinstance(layers, str):
        layers = [layers]
    if not isinstance(sites, bool) and not isinstance(sites, list):
        sites = [sites]

    # Retrieve metadata
    yaxis_label = yaxis_label or fields[0].metadata['name']
    unit = unit or fields[0].metadata['units']
    layers = layers or fields[0].layers

    # Check valid labels
    if len(fields) > 1 and field_labels is None:
            raise ValueError("field_labels must be defined when there are more than one field.")
    
    # If only one field and no field_labels provided, don't use label
    if len(fields) == 1 and field_labels is None:
        field_labels = " "

    # Convert all series to DataFrames with flattened structure
    combined_data = []
    for i, field in enumerate(fields):
        label = field_labels[i]
        if gridop: # Reduce grid if gridop is provided
            field_red = field.red_space(gridop)
        df = cp.make_lines(field_red if gridop else field)

        if sites: # If sites, flatten data
            df = select_sites(df, sites=sites)
            layers = list(df.columns) # Update layers

        for layer in layers:
            data = df[layer].values.flatten()
            combined_data.append(pd.DataFrame({
                "value": data,
                "series": label,
                "layer": layer
            }))

    df_long = pd.concat(combined_data, ignore_index=True)

    # Get color palette - one color per field/series (not per layer)
    n_classes = len(field_labels)
    colors, color_dict = get_color_palette(n_classes=n_classes, palette=palette, custom_palette=custom_palette)
    palette = {label: color for label, color in zip(field_labels, colors)}

    # Set axes depending on orientation
    x, y = ("value", "layer") if horizontal else ("layer", "value")

    # Base arguments
    plot_kwargs = {
        "data": df_long,
        "x": x,
        "y": y,
        "hue": "series",
        "kind": plot_type,
        "palette": palette,
        "height": y_fig,
        "aspect": x_fig / y_fig
    }


    plot_kwargs.update(kwargs)

    # Recommended arguments
    if plot_type == "box" or plot_type == "boxen":
        plot_kwargs["fill"] = False
        plot_kwargs["showfliers"] = False
        plot_kwargs["gap"] = 0.1
    if plot_type == "violin":
        plot_kwargs["inner"] = None
        plot_kwargs["bw_method"] = 1
        if len(fields) == 2:
            plot_kwargs["split"] = True

    # Create the catplot
    fig = sns.catplot(**plot_kwargs)

    ax = fig.ax  # Extract the main axis

    # Set labels and title
    axis_label = f"{yaxis_label} (in {unit})" if unit != "[no units]" else yaxis_label
    if horizontal:
        ax.set_xlabel(axis_label, fontsize=16)
        ax.set_ylabel("")
    else:
        ax.set_ylabel(axis_label, fontsize=16)
        ax.set_xlabel("")

    ax.tick_params(labelsize=14)
    if title:
        ax.set_title(title, fontsize=16)

    # Rotate labels if requested to prevent overlap
    if vertical_xlabels:
        if horizontal:
            # When horizontal, rotate y-axis labels (which show the layer names)
            ax.set_yticklabels(ax.get_yticklabels(), rotation=90, ha='center')
        else:
            # When vertical, rotate x-axis labels (which show the layer names)
            ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha='right')

    # Custom legend (colored labels, no box)
    fig._legend.remove()
    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles=[plt.Line2D([], [], color=palette[label], marker='', linestyle='') for label in labels],
              labels=labels, handlelength=0, handletextpad=0, labelcolor=[palette[label] for label in labels],
              loc='best', frameon=False, fontsize=14)

    # Apply dark mode
    if dark_mode:
        fig, ax = make_dark_mode(fig.fig, ax)

    return handle_figure_output(fig, output_file=output_file, return_fig=return_fig, transparent=transparent)



# -----------------------------------------------------------------------------
# | Suite of unit tests for spatial reduction operations on the 'lonlat' grid |
# -----------------------------------------------------------------------------

import numpy as np
import pandas as pd
import pytest
import canopy.core.frameops as frameops
from canopy.core.constants import *
from canopy import Field
import xarray as xr

INDEX_NAMES = ['lon', 'lat', 'time', ]

# The test fields must follow the Field specification, i.e.;
# - field._data is a pandas DataFrame with multiindex with levels 'lon', 'lat', 'time'
# - field._grid must be consistent with the data
TEST_DATA = {
        # Annual npp over Spain 1990-2010, gridded
        'anpp_spain': 'tests/test_data/anpp_spain_1990_2010.out.gz',
        }

def get_test_field():
    # Change this to use another test dataframe
    test = 'anpp_spain'
    return Field.from_file(TEST_DATA[test])


def test_red_space_av_lat():
    """Check that lat-mean yields same result as when calc'd with xarray"""
    field = get_test_field()
    df, grid = field._data, field._grid
    # PyGUESS calculation
    # -------------------
    red_df, _, _ = frameops.red_space(df, grid, 'av', 'lat')
    # Drop 'lon' index before converting to xarray for comparison
    red_df.index = red_df.index.droplevel('lat')
    red_df = red_df.to_xarray()
    # Xarray calculation
    # ------------------
    xa = df.to_xarray().sortby('lon').sortby('lat').sortby('time')
    red_xa = xa.mean('lat')
    # Comparison
    # ----------
    xr.testing.assert_allclose(red_xa, red_df)

def test_red_space_av_lon():
    """Check that lon-mean yields same result as when calc'd with xarray"""
    field = get_test_field()
    df, grid = field._data, field._grid
    # PyGUESS calculation
    # -------------------
    red_df, _, _ = frameops.red_space(df, grid, 'av', 'lon')
    # Drop 'lon' index before converting to xarray for comparison
    red_df.index = red_df.index.droplevel('lon')
    red_df = red_df.to_xarray()
    # Xarray calculation
    # ------------------
    xa = df.to_xarray().sortby('lon').sortby('lat').sortby('time')
    red_xa = xa.mean('lon')
    # Compare them
    xr.testing.assert_allclose(red_xa, red_df)

def test_red_space_av_both():
    """Check that both-mean yields same result as when calc'd with xarray"""
    field = get_test_field()
    df, grid = field._data, field._grid
    # PyGUESS calculation
    # -------------------
    red_df, grid, _ = frameops.red_space(df, grid, 'av', 'both')
    # Drop 'lon' and 'lat' indices before converting to xarray for comparison
    red_df.index = red_df.index.droplevel(['lon','lat'])
    red_df = red_df.to_xarray()
    # Xarray calculation
    # ------------------
    xa = df.to_xarray().sortby('lon').sortby('lat').sortby('time')
    weights = np.cos(np.deg2rad(xa.lat))
    xa_weighted = xa.weighted(weights)
    red_xa = xa_weighted.mean(('lon', 'lat'))
    # Compare them
    xr.testing.assert_allclose(red_xa, red_df)

def test_red_space_sum_lat():
    """Check that lat-sum yields same result as when calc'd with xarray"""
    field = get_test_field()
    df, grid = field._data, field._grid
    # PyGUESS calculation
    # -------------------
    red_df, _, _ = frameops.red_space(df, grid, 'sum', 'lat')
    # Drop 'lat' index before converting to xarray for comparison
    red_df.index = red_df.index.droplevel('lat')
    red_df = red_df.to_xarray()
    # Xarray calculation
    # ------------------
    xa = df.to_xarray().sortby('lon').sortby('lat').sortby('time')
    red_xa = (xa*EARTH_RADIUS*np.deg2rad(grid.dx['lat'])).sum('lat')
    # Comparison
    # ----------
    xr.testing.assert_allclose(red_xa, red_df)


def test_red_space_sum_lon():
    """Check that lon-sum yields same result as when calc'd with xarray"""
    field = get_test_field()
    df, grid = field._data, field._grid
    # PyGUESS calculation
    # -------------------
    red_df, _, _ = frameops.red_space(df, grid, 'sum', 'lon')
    # Drop 'lon' index before converting to xarray for comparison
    red_df.index = red_df.index.droplevel('lon')
    red_df = red_df.to_xarray()
    # Xarray calculation
    # ------------------
    xa = df.to_xarray().sortby('lon').sortby('lat').sortby('time')
    weights = EARTH_RADIUS*np.deg2rad(grid.dx['lon'])*np.cos(np.deg2rad(xa.lat))
    xa_weighted = xa.weighted(weights)
    red_xa = xa_weighted.sum('lon')
    # Comparison
    # ----------
    xr.testing.assert_allclose(red_xa, red_df)


def test_red_space_sum_both():
    """Check that both-sum yields same result as when calc'd with xarray"""
    field = get_test_field()
    df, grid = field._data, field._grid
    # PyGUESS calculation
    # -------------------
    red_df, _, _ = frameops.red_space(df, grid, 'sum', 'both')
    # Drop 'lon' and 'lat' indices before converting to xarray for comparison
    red_df.index = red_df.index.droplevel(['lon','lat'])
    red_df = red_df.to_xarray()
    # Xarray calculation
    # ------------------
    xa = df.to_xarray().sortby('lon').sortby('lat').sortby('time')
    weights = EARTH_RADIUS**2*np.deg2rad(grid.dx['lon'])*np.deg2rad(grid.dx['lat'])*np.cos(np.deg2rad(xa.lat))
    xa_weighted = xa.weighted(weights)
    red_xa = xa_weighted.sum(('lon', 'lat'))
    # Compare them
    xr.testing.assert_allclose(red_xa, red_df)


def test_levels_av_lon():
    """Check that the data frame index still has all levels after averaging in longitude"""
    field = get_test_field()
    df, grid = field._data, field._grid
    red_df, _, _ = frameops.red_space(df, grid, 'av', 'lon')
    assert red_df.index.names == INDEX_NAMES


def test_levels_av_lat():
    """Check that the data frame index still has all levels after averaging in latitude"""
    field = get_test_field()
    df, grid = field._data, field._grid
    red_df, _, _ = frameops.red_space(df, grid, 'av', 'lat')
    assert red_df.index.names == INDEX_NAMES

def test_levels_av_both():
    """Check that the data frame index still has all levels after averaging on both axes"""
    field = get_test_field()
    df, grid = field._data, field._grid
    red_df, _, _ = frameops.red_space(df, grid, 'av', 'both')
    assert red_df.index.names == INDEX_NAMES

def test_levels_sum_lon():
    """Check that the data frame index still has all levels after aggregating in longitude"""
    field = get_test_field()
    df, grid = field._data, field._grid
    red_df, _, _ = frameops.red_space(df, grid, 'sum', 'lon')
    assert red_df.index.names == INDEX_NAMES

def test_levels_sum_lat():
    """Check that the data frame index still has all levels after aggregating in latitude"""
    field = get_test_field()
    df, grid = field._data, field._grid
    red_df, _, _ = frameops.red_space(df, grid, 'sum', 'lat')
    assert red_df.index.names == INDEX_NAMES

def test_levels_sum_both():
    """Check that the data frame index still has all levels after aggregating on both axes"""
    field = get_test_field()
    df, grid = field._data, field._grid
    red_df, _, _ = frameops.red_space(df, grid, 'sum', 'both')
    assert red_df.index.names == INDEX_NAMES




from typing_extensions import Self
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import pandas as pd

class Grid(ABC):
    """Abstract base class for a generic Grid object

    A grid object describes the grid associated to the data in a Field object.
    """

    def __init__(self,
                 grid_type: str,
                 axis0: str, axis1: str,
                 gridop0: str | None = None, gridop1: str | None = None):
        """The Grid superclass constructor

        Parameters
        ----------
        axis0
            The name of the first axis (e.g. 'x' or 'lon').
        axis1
            The name of the second axis (e.g. 'y' or 'lat').
        """
        self._grid_type = grid_type
        self._axis_names: tuple = (axis0, axis1)
        self._axis_gridop: dict['str','str' | None] = {name: gridop for name, gridop in zip(self.axis_names, [gridop0, gridop1])}


    @property
    def grid_type(self):
        return self._grid_type

    @property
    def axis_names(self):
        return self._axis_names

    @property
    def axis_gridop(self):
        return self._axis_gridop

    @classmethod
    @abstractmethod
    def from_frame(cls, df: pd.DataFrame) -> Self:
        """Create a grid from a DataFrame.

        The method should construct a grid object of the subclassed type with the information
        from the pandas DataFrame. For an example, see the Grid subclasses defined in the grid/
        folder (files grid_sites.py, ...)

        Parameters
        ----------
        df
            A pandas DataFrame with a valid format (see Field documentation).

        Returns
        -------
        An instance of the grid subclass.
        """
        pass

    @abstractmethod
    def to_txt(self, fname: str, sep: str = ' ', randomize: bool = False) -> None:
        """Save a list of grid points as a text file

        Parameters
        ----------
        fname: str
            The file name to save the gridlist to.
        sep: str
            Separator between the two coordinates.
        randomize: bool
            Whether to shuffle the list before saving it.
        """
        pass

    @abstractmethod
    def get_sliced_grid(self, axis0_slice: tuple[float,float] | None, axis1_slice: tuple[float,float] | None) -> 'Grid':
        """Create a new grid, sliced accordingto the parameters

        Parameters
        ----------
        axis0_slice
            Specifies an interval on axis0.
        axis1_slice
            Specifies an interval on axis1.

        Returns
        -------
        An instance of the grid subclass.
        """
        pass

    @abstractmethod
    def get_reduced_grid(self, gridop: str, axis: str) -> 'Grid':
        """Create a new grid, reduced according to the parameters

        Parameters
        ----------
        gridop
            The reduction operation
        axis
            The axis to be reduced

        Returns
        -------
        An instance of the grid subclass

        Notes
        -----
        This function has two responsibilities:
            To change the value of `self.axis_gridop`:
                self.axis_gridop[axis] = gridop
                
            To 'collapse' the reduced axis. For example, if the reduction
            is along the 'x' axis, replace the representation of this axis by
            some placehoder value, like `None` or `np.empty[]`. The concrete
            implementation will depend on how axes are represented in the specific
            Grid subclass.
        """
        pass

    @abstractmethod
    def is_compatible(self, other: Self) -> bool:
        """ Check wether grid is compatible with another grid instance.

        Two grids are compatible if the associated mathematical grids are the same.

        To illustrate this point, consider two lonlat Grid objects with the same coordinate increments.
        The latitude of a point in the first Grid object is 17.5, and the latitude of a point in
        the second Grid object is 3.25. These two grids are not compatible, because the mathematical grid
        associated to the first object cannot contain the second point.

        Notes
        -----
        Compatible Grid objects can represent overlapping regions.

        Site grids are always equivalent to each other.

        Paramaters
        ----------
        other
            A grid to compare the calling object with.

        Returns
        -------
        True if the grids are compatible, False otherwise.
        """
        pass

    def __repr__(self) -> str:
        return f"Grid type: {self.grid_type}"

    def __str__(self) -> str:
        return self.__repr__()

    @abstractmethod
    def __add__(self, other):
        pass

    def is_reduced(self, axis: str) -> bool:
        """Check whether axis has been reduced.

        Parameters
        ----------
        axis
            The name or index of the axis to check.

        Returns
        -------
            A boolean: True if the axis has been reduced, False otherwise.
        """
        if isinstance(axis, int):
            axis = self.axis_names[axis]

        return self.axis_gridop[axis] is not None






"""
Utilities for site-based fields
"""
import numpy as np
import pandas as pd
from canopy.core.field import Field
from canopy.grid import get_grid_type, get_grid


def check_grid_is_sites(field: Field):
    if get_grid_type(field.grid) != 'sites':
        raise ValueError(f"Grid must be of type 'sites' (passed: {get_grid_type(field.grid)}).")


def sel_sites(field: Field, sites: list[tuple[int | float, int | float]], invert: bool = False) -> Field:
    """
    Select sites from a Field

    Parameters
    ----------
    field : Field
        The field to select the sites from
    sites : list[tuple[int | float, int | float]]
        A list of coordinates
    invert : bool
        Invert selection: select sites not in the 'sites' list

    Returns
    -------
    A new Field with the selected sites
    """

    # TODO: select sites by NAME -> GridSites should include names
    check_grid_is_sites(field)

    ax0 = field.data.index.get_level_values(0)
    ax1 = field.data.index.get_level_values(1)

    spatial_index = pd.MultiIndex.from_arrays([ax0, ax1])
    mask = spatial_index.isin(sites)
    if invert:
        mask = np.logical_not(mask)

    data = field.data.loc[mask]
    if data.empty:
        grid = get_grid('empty').from_frame(data)
    else:
        grid = get_grid('sites').from_frame(data)

    field_new = Field(data, grid)

    return field_new



from canopy.core.field import Field
from canopy.sources.source_abc import Source
from canopy.sources.registry import register_source
from canopy.source_data import get_source_data
from types import MappingProxyType
import itertools
import warnings

FREQS = {'HH': 'Half-hourly', 'HR': 'Hourly', 'DD': 'Daily', 'WW': 'Weekly', 'MM': 'Monthly', 'YY': 'Yearly', }

FIELDS = {
    'H': 'Sensible Heat Flux',
    'LE': 'Latent Heat Flux',
    'G': 'Ground Surface Heat Flux',
    'NEE': 'Net Ecosystem Exchange',
    'RECO': 'Ecosystem Respiration',
    'GPP': 'Gross Primary Producion',
}

FIELD_COLS = {
    'H': ('H_F_MDS', 'H_F_MDS_QC', 'H_CORR', 'H_CORR_25', 'H_CORR_75', 'H_RANDUNC'),
    'H_longfreq': ('H_F_MDS', 'H_F_MDS_QC', 'H_CORR', 'H_RANDUNC'),
    'LE': ('LE_F_MDS', 'LE_F_MDS_QC', 'LE_CORR', 'LE_CORR_25', 'LE_CORR_75', 'LE_RANDUNC'),
    'LE_longfreq': ('LE_F_MDS', 'LE_F_MDS_QC', 'LE_CORR', 'LE_RANDUNC'),
    'G': ('G_F_MDS', 'G_F_MDS_QC'),
    'NEE': (
        'NEE_VUT_REF', 'NEE_VUT_REF_QC', 'NEE_VUT_REF_RANDUNC',
        'NEE_VUT_05', 'NEE_VUT_16', 'NEE_VUT_25', 'NEE_VUT_50', 'NEE_VUT_75', 'NEE_VUT_84', 'NEE_VUT_95', 
        'NEE_VUT_05_QC', 'NEE_VUT_16_QC', 'NEE_VUT_25_QC', 'NEE_VUT_50_QC', 'NEE_VUT_75_QC', 'NEE_VUT_84_QC', 'NEE_VUT_95_QC', 
    ),
    'RECO': (
        'RECO_NT_VUT_REF', 'RECO_DT_VUT_REF',
        'RECO_NT_VUT_05', 'RECO_NT_VUT_16', 'RECO_NT_VUT_25', 'RECO_NT_VUT_50', 'RECO_NT_VUT_75', 'RECO_NT_VUT_84', 'RECO_NT_VUT_95', 
        'RECO_DT_VUT_05', 'RECO_DT_VUT_16', 'RECO_DT_VUT_25', 'RECO_DT_VUT_50', 'RECO_DT_VUT_75', 'RECO_DT_VUT_84', 'RECO_DT_VUT_95', 
        'RECO_SR', 'RECO_SR_N',
    ),
    'GPP': (
        'GPP_NT_VUT_REF', 'GPP_DT_VUT_REF',
        'GPP_NT_VUT_05', 'GPP_NT_VUT_16', 'GPP_NT_VUT_25', 'GPP_NT_VUT_50', 'GPP_NT_VUT_75', 'GPP_NT_VUT_84', 'GPP_NT_VUT_95', 
        'GPP_DT_VUT_05', 'GPP_DT_VUT_16', 'GPP_DT_VUT_25', 'GPP_DT_VUT_50', 'GPP_DT_VUT_75', 'GPP_DT_VUT_84', 'GPP_DT_VUT_95', 
    )
}


def _get_field_cols(field_name: str, freq: str):
    if field_name in ['H', 'LE'] and freq in ['WW', 'MM', 'YY']:
        field_name += '_longfreq'
    return FIELD_COLS[field_name]


def _select_fluxnet_sites(site_codes: list, source_data: dict):

    valid_sites = []
    invalid_sites = []
    for sc in site_codes:
        if sc in source_data['sites']:
            valid_sites.append(sc)
        else:
            invalid_sites.append(sc)

    return valid_sites, invalid_sites


@register_source('fluxnet2015')
class SourceFluxnet2015(Source):
    """
    Source object for Fluxnet2015 data
    """

    def __init__(self, path, sites=None) -> None:
        super().__init__(path, get_source_data('fluxnet2015'))

        # TODO: possible duplication with file reader fuction (readers.fluxnet2015())
        site_codes_directory = [fname.name[4:10].lower() for fname in self.path.glob(f"FLX_*FLUXNET2015*.zip")]

        valid_sites_directory, invalid_sites_directory = _select_fluxnet_sites(site_codes_directory, self.source_data)
        if len(invalid_sites_directory) > 0:
            warnings.warn(f"The following site files were found in '{path}', but are not valid FLUXNET2015 sites: {invalid_sites_directory}.")
        if len(valid_sites_directory) == 0:
            raise ValueError(f"No valid FLUXNET2015 files found in '{path}'.")

        if sites is None:
            site_list = valid_sites_directory
        else:
            valid_sites_passed, invalid_sites_passed = _select_fluxnet_sites(sites, self.source_data)

            if len(invalid_sites_passed) > 0:
                warnings.warn(f"The following sites are not valid FLUXNET2015 sites {invalid_sites_passed}.")
            if len(valid_sites_passed) == 0:
                raise ValueError(f"No valid FLUXNET2015 files passed.")
            
            vset_directory = set(valid_sites_directory)
            vset_passed = set(valid_sites_passed)
            not_in_directory = list(vset_passed - vset_directory)
            if len(not_in_directory) > 0:
                warnings.warn(f"No file found in '{path}' for the following sites: {not_in_directory}.")

            site_list = list(vset_directory & vset_passed)
            if len(site_list) == 0:
                raise ValueError(f"No files for any of the selected sites found in '{path}'.")
            
        self._sites = {}
        for site in sorted(site_list):
            self._sites[site] = (self.source_data['sites'][site]['longitude'], self.source_data['sites'][site]['latitude'])

        for field_name, freq in itertools.product(FIELDS, FREQS):
            field_id = f"{field_name}_{freq}"
            self._fields[field_id] = None
            self.is_loaded[field_id] = False
    

    @property
    def sites(self):
        return MappingProxyType(self._sites)


    def load_field(self, field_id: str) -> Field:

        field_id = field_id.upper()
        if field_id not in self.fields:
            raise KeyError(f"Field '{field_id}' not found in source.")
        field_name, freq = field_id.split('_')

        field_cols = _get_field_cols(field_name, freq)
        field = Field.from_file(self.path,
                                file_format='fluxnet2015',
                                grid_type='sites',
                                cols=field_cols,
                                freq=freq,
                                sites=self.sites)

        field.add_md('source', self.source)
        field.set_md('name', self.source_data['fields'][field_id]['name'])
        field.set_md('description', self.source_data['fields'][field_id]['description'])
        field.set_md('units', self.source_data['fields'][field_id]['units'])

        self.is_loaded[field_id] = True
        self._fields[field_id] = field

        name_dict = {}
        # The constructor ensures that all sites in self.sites exist in the source
        for site in self.sites:
            name_dict[(self.source_data['sites'][site]['longitude'], self.source_data['sites'][site]['latitude'])] = site
        field.grid.assign_names(name_dict)

        return field


    def get_gridlist(self, fname: str = None) -> str | None:

        gridlist = []
        for site in self.sites:
            gridlist.append(f"{self.source_data['sites'][site]['longitude']} {self.source_data['sites'][site]['latitude']} {site}")

        if fname is not None:
            with open(fname, 'w') as f:
                f.write("\n".join(gridlist))
        else:
            return "\n".join(gridlist)




import pandas as pd
import numpy as np
from pathlib import Path
from zipfile import ZipFile
from datetime import datetime
from canopy.readers.registry import register_reader_desc
from canopy.source_data import get_source_data

FLUXNET_FREQS = ['HH', 'HR', 'DD', 'WW', 'MM', 'YY', ]
PERIOD_FREQS = {'HH': '30min', 'HR': 'h', 'DD': 'D', 'WW': 'W', 'MM': 'M', 'YY': 'Y', }
TIMESTAMP_FORMAT = {'HH': '%Y%m%d%H%M', 'HR': '%Y%m%d%H%M', 'DD': '%Y%m%d', 'WW': '%Y%m%d', 'MM': '%Y%m', 'YY': '%Y', }

NA = -9999.0

def _read_zipped_fnet(path: Path, freq: str, subset: str, cols: list[str] | None):

    csv_file = Path(path.stem.replace(f"{subset}_", f"{subset}_{freq}_") + ".csv")

    if not freq in FLUXNET_FREQS:
        raise ValueError(f"Unrecognized frequency {freq} (must be one of {FLUXNET_FREQS}).")
    timestamp_col = 'TIMESTAMP'
    if freq in ['HH', 'HR', 'WW']:
        timestamp_col += '_START'
    timestamp_format = TIMESTAMP_FORMAT[freq]

    with ZipFile(path) as fzip:
        try:
            fzip.extract(csv_file.name)
        except KeyError:
            return None

    with open(csv_file) as f:
        csv_header = f.readline().split(',')
        csv_header = [c.strip() for c in csv_header]
        for c in ['TIMESTAMP', 'TIMESTAMP_START', 'TIMESTAMP_END']:
            try:
                csv_header.remove(c)
            except ValueError:
                pass

    if cols is None:
        cols_filtered = [timestamp_col] + csv_header
    else:
        cols = [c.strip().upper() for c in cols]
        cols_filtered = sorted(list(set(cols) & set(csv_header)))
        cols_filtered = [timestamp_col] + cols_filtered

    site_code = csv_file.name[4:10].lower()
    fnet_data = get_source_data('fluxnet2015')
    lon = fnet_data["sites"][site_code]['longitude']
    lat = fnet_data["sites"][site_code]['latitude']

    df = pd.read_csv(csv_file,
                     usecols=cols_filtered,
                     parse_dates = [timestamp_col],
                     date_format = timestamp_format)
        
    csv_file.unlink()
    df.rename(columns={timestamp_col:'time'}, inplace=True)
    df['time'] = df['time'].dt.to_period(PERIOD_FREQS[freq])
    df['lon'] = lon
    df['lat'] = lat
    df.index = pd.MultiIndex.from_frame(df[['lon', 'lat', 'time']])
    df.drop(['lon', 'lat', 'time'], axis=1, inplace=True)
    df[df == NA] = np.nan

    return df


@register_reader_desc('FLUXNET 2015')
def fluxnet2015(path: str | Path,
                grid_type: str,
                freq: str,
                subset: str = 'SUBSET',
                cols: list[str] | None = None,
                sites: list[str] | None = None):
    """
    Read a file from a .zip archive from FLUXNET2015 as downloaded from www.fluxnet.org

    Parameters
    ----------
    path : str | Path
        The path of a fluxnet.org .zip file or the path of a directory containing fluxnet.org
        .zip files. In the latter case, see 'sites' argument below.
    freq : str
        The frequency to retrieve as described in fluxnet.org:
            HH: half-hourly
            HR: hourly
            DD: daily
            WW: weekly
            MM: monthly
            YY: yearly
    subset : str
        The subset to retrieve. One of 'FULLSET' or 'SUBSET' (the default).
    cols : list[str] or None
        Colums to read (case insensitive). If None, the default, all columns are read.
    sites : list[str] or None
        A list of fluxnet site codes as in https://fluxnet.org/sites/site-list-and-pages/
        If None, all valid fluxnet sites in the path are read.
    """

    path = Path(path)

    if not path.exists():
        raise ValueError(f"Path {path} does not exist.")

    # If path is a file, attempt to read that file
    if path.is_file():
        df = _read_zipped_fnet(path, freq, subset, cols)
    # If path is a directory, load all files in directory, or all files that are in "sites"
    else:
        dframes = []
        if sites is None:
            files = path.glob("*FLUXNET2015*.zip")
            for file in files:
                dframes.append(_read_zipped_fnet(file, freq, subset, cols))
        else:
            match = set()
            for site_code in sites:
                for file in path.glob(f"FLX_*FLUXNET2015*{subset}*.zip"):
                    if site_code.lower() in file.name.lower():
                        dframes.append(_read_zipped_fnet(file, freq, subset, cols))
                        match.add(site_code)
                        break
            if len(dframes) == 0:
                raise ValueError(f"No matching files for 'sites' list in {path}.")
            no_match = set(sites) ^ match
            if len(no_match) > 0:
                print("WARNING: no matching files for the following sites:")
                print('\n'.join(no_match))
        try:
            df = pd.concat(dframes)
        except ValueError:
            df = None

    if df is None:
        raise ValueError("No matching data found.")

    if not df.index.is_monotonic_increasing:
        df.sort_index(inplace=True)

    return df




import json
from canopy.json.json_registry import get_json_function
import difflib

EXPECTED_KEYS = {
    'figure', 'output_file', 'input_file', 'grid_type', 'time_slice', 'lat_slice', 'lon_slice',
    'drop_layers', 'red_layers', 'layer', 'categorical', 'redop', 'timeop', 'gridop',
    'cb_label', 'title', 'scatter_size', 'scatter_alpha',
    'unit', 'n_classes', 'classification', 'palette', 'custom_palette', 'orientation',
    'extend', 'proj', 'force_zero', 'dark_mode', 'transparent', 'x_fig', 'y_fig',
    'percentage', 'layers', 'make_diff', 'yaxis_label', 'field_labels', 'move_legend',
    'legend_style', 'reverse_hue_style', 'max_labels_per_col', 'baseline', 'rolling_size',
    'stacked', 'relative', 'no_style', 'field_a_label', 'field_b_label', 'unit_a',
    'unit_b', 'plot_type', 'field_labels', 'horizontal', 'vertical_xlabels' 'kwargs',
    'stats_annotation', 'nonsig', 'source', 'convert_units', 'filter', 'yaxis_lim',
    'merge_fields', 'merge_interval', 'region', 'region_type'
}

def run_json(config_file):
    """
    Run a function based on the JSON configuration file.

    Parameters
    ----------
    config_file : str
        Path to the JSON configuration file.

    Notes
    -----
    This function loads the configuration from the specified JSON file,
    retrieves the appropriate function from the registry, and executes it
    with the configuration as an argument.
    """

    # Load the configuration from the JSON file
    json_config = load_json_config(config_file)

    # Check if any JSON keys are not exepcted
    warn_unused_keys(json_config, EXPECTED_KEYS)

    # Get the appropriate function from the registry and call it
    json_function = get_json_function(json_config)
    json_function(json_config)

def load_json_config(file_path):
    with open(file_path) as f:
        return json.load(f)

def warn_unused_keys(json_config, expected_keys):
    for key in json_config:
        if key not in expected_keys:
            suggestion = difflib.get_close_matches(key, expected_keys, n=1)
            if suggestion:
                print(f"Warning: The parameter '{key}' is not recognized. Did you mean '{suggestion[0]}'?")
            else:
                print(f"Warning: The parameter '{key}' in your JSON will not be used.")



# -----------------------------------------
# | Tests for map generation using canopy |
# -----------------------------------------

import pytest
from pathlib import Path
import matplotlib
matplotlib.use('Agg')  # Set non-GUI backend before pyplot import
from canopy import Field
import canopy.visualization as cv

DATA_DIR = Path("tests/test_data")

@pytest.mark.filterwarnings("ignore:FigureCanvasAgg is non-interactive, and thus cannot be shown:UserWarning")
def test_make_simple_map():
    """Test make_simple_map on example field without error"""

    field = Field.from_file(DATA_DIR / "anpp_spain_1990_2010.out.gz")

    cv.make_simple_map(
        field=field,
        layer="Total",
        timeop="sum",
        cb_label="Carbon Pool",
        unit="kgC m-2",
        title="Carbon Pool in Spain from 1990 to 2010 (summed over time)",
        n_classes=3,
        classification=[0, 10, 20, 40],
        palette="YlOrRd",
        custom_palette=None,
        extend="both",
        proj="EuroPP",
        force_zero=True,
        dark_mode=True,
        transparent=True,
        x_fig=10,
        y_fig=8
    )

@pytest.mark.filterwarnings("ignore:FigureCanvasAgg is non-interactive, and thus cannot be shown:UserWarning")
def test_make_diff_map():
    """Test make_simple_map on example field without error"""

    field_a = Field.from_file(DATA_DIR / "anpp_spain_1990_2010.out.gz")
    field_b = Field.from_file(DATA_DIR / "anpp_spain_1990_2010_mod.out.gz")

    cv.make_diff_map(
        field_a=field_a,
        field_b=field_b,
        layer="Total",
        timeop="sum",
        cb_label="Carbon Pool",
        unit="kgC m-2",
        title="Carbon Pool in Spain from 1990 to 2010 (summed over time)",
        n_classes=6,
        classification="linear",
        palette="YlOrRd",
        custom_palette=None,
        extend="both",
        proj="EuroPP",
        force_zero=True,
        dark_mode=True,
        transparent=True,
        x_fig=10,
        y_fig=8
    )


import numpy as np
from dataclasses import dataclass

@dataclass
class Raster:
    xx: np.ndarray | None = None
    yy: np.ndarray | None = None
    vmap: np.ndarray | None = None
    labels: dict | None = None




try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"

def version():
    return __version__

from canopy.core.field import Field
from canopy.core.raster import Raster
from canopy.core.redspec import RedSpec
from canopy.core.constants import *
from canopy.sources import get_source
from canopy.util.fieldops import make_raster, make_lines, concat2, merge_fields, filter_region
from canopy.util.compare_ts import compare_ts
from canopy.json.run_json import run_json
from canopy.tests.test_data.registry import get_test_data



"""Grid with longitude and latitude coordinates.

This grid type has longitude and latutide coordinates. The longitude and latitude intervals are
constant, although they can be different (i.e. dlon != dlat). Reduction operations on this grid
are weighted according to their position on the grid.
"""
import pandas as pd
import numpy as np
import copy
from typing_extensions import Self
from canopy.grid.grid_abc import Grid
from canopy.grid.registry import register_grid, register_gridop, get_gridop
from canopy.core.constants import *


grid_type = 'lonlat'

# ---------------
# GRID DEFINITION
# ---------------

COORD_BOUNDARIES = {'lon': -180., 'lat': -90.}

@register_grid(grid_type)
class GridLonLat(Grid):
    """A Grid type representing a longitude-latitude grid with constant increments."""

    def __init__(self,
                 lon_min: float = np.nan, lon_max: float = np.nan, dlon: float = np.nan,
                 lat_min: float = np.nan, lat_max: float = np.nan, dlat: float = np.nan,
                 lon_gridop: str | None = None, lat_gridop: str | None = None) -> None:
        """
        Parameters
        ----------
        lon_min : float
            Smallest longitude of the represented domain (degrees).
        lon_max : float
            Largest longitude of the represented domain (degrees).
        dlon : float
            Increment in longitude (degrees). 
        lat_min : float
            Smallest latitude of the represented domain (degrees).
        lat_max : float
            Largest latitude of the represented domain (degrees).
        dlat : float
            Increment in latitude (degrees). 
        lon_gridop : str | None
            Grid reduction operation along lon axis. If not None, values of lon_min, lon_max, and dlon are ignored.
        lat_gridop : str | None
            Grid reduction operation along lat axis. If not None, values of lat_min, lat_max, and dlat are ignored.
        """

        if lon_gridop is not None:
            try:
                get_gridop(grid_type, lon_gridop, 'lon')
            except:
                raise ValueError(f"Operation '{lon_gridop}' not registered for axis 'lon'")
        if lat_gridop is not None:
            try:
                get_gridop(grid_type, lat_gridop, 'lat')
            except:
                raise ValueError(f"Operation '{lat_gridop}' not registered for axis 'lat'")

        # TODO: this to ABC?
        if lon_gridop is not None and lat_gridop is not None and lon_gridop != lat_gridop:
            raise ValueError(f"Grid operation must be the same on both axes")

        super().__init__(grid_type, axis0='lon', axis1='lat', gridop0=lon_gridop, gridop1=lat_gridop)
        self.axis_gridop['both'] = None

        self.xmin = {}
        self.xmax = {}
        self.dx = {}
        self.axes = {}

        param_xmin = {'lon': lon_min, 'lat': lat_min}
        param_xmax = {'lon': lon_max, 'lat': lat_max}
        param_dx = {'lon': dlon, 'lat': dlat}
        param_gridop = {'lon': lon_gridop, 'lat': lat_gridop}

        for axis_name in self.axis_names:
            if param_gridop[axis_name] is None:
                if param_dx[axis_name] <= 0:
                    raise ValueError(f"d{axis_name} must be greater than 0.")
                if param_xmin[axis_name] > param_xmax[axis_name]:
                    raise ValueError(f"{axis_name}_min must be smaller or equal than {axis_name}_max.")
                self.axes[axis_name] = np.arange(param_xmin[axis_name],
                                                 param_xmax[axis_name]+0.5*param_dx[axis_name],
                                                 param_dx[axis_name])
                self.xmin[axis_name] = param_xmin[axis_name]
                self.xmax[axis_name] = param_xmax[axis_name]
                self.dx[axis_name] = param_dx[axis_name]
            else:
                self.axes[axis_name] = np.empty(0)
                self.xmin[axis_name] = np.nan
                self.xmax[axis_name] = np.nan
                self.dx[axis_name] = np.nan


    def to_txt(self, fname: str, sep: str = ' ', randomize: bool = False) -> None:
        raise NotImplementedError


    def get_reduced_grid(self, gridop: str, axis: str) -> Grid:
        """Create a new grid, reduced according to the parameters

        Parameters
        ----------
        gridop : str
            The reduction operation
        axis : str
            The axis to be reduced

        Returns
        -------
        An instance of GridLonLat
        """
        grid = copy.deepcopy(self)
        for an in grid.axis_names:
            if axis == an or axis == 'both':
                grid.axis_gridop[an] = gridop
                grid.axes[an] = np.empty(0)
                grid.dx[an] = np.nan
                grid.xmin[an] = np.nan
                grid.xmax[an] = np.nan

        return grid


    def get_sliced_grid(self, axis0_slice: tuple[float,float] | None, axis1_slice: tuple[float,float] | None) -> Grid:
        """Create a new grid, sliced according to the parameters.

        Parameters
        ----------
        axis0_slice : tuple[float, float]
            Specifies an interval on longitude.
        axis1_slice : tuple[float, float]
            Specifies an interval on latitude.

        Returns
        -------
        An instance of the grid subclass.
        """
        xmin = {ax: np.nan for ax in self.axis_names}
        xmax = {ax: np.nan for ax in self.axis_names}
        dx = {ax: np.nan for ax in self.axis_names}
        gridop = {ax: None for ax in self.axis_names}
        for axis_name, axis_slice in zip(self.axis_names, [axis0_slice, axis1_slice]):
            if self.is_reduced(axis_name):
                gridop[axis_name] = self.axis_gridop[axis_name]
                continue
            if axis_slice is None:
                xmin[axis_name] = self.xmin[axis_name]
                xmax[axis_name] = self.xmax[axis_name]
                dx[axis_name] = self.dx[axis_name]
            else:
                xmin[axis_name], xmax[axis_name] = axis_slice
                xmin[axis_name] = self.axes[axis_name][self.axes[axis_name] >= xmin[axis_name]].min()
                xmax[axis_name] = self.axes[axis_name][self.axes[axis_name] <= xmax[axis_name]].max()
                dx[axis_name] = self.dx[axis_name]

        return GridLonLat(lon_min = xmin['lon'], lon_max = xmax['lon'], dlon = dx['lon'],
                          lat_min = xmin['lat'], lat_max = xmax['lat'], dlat = dx['lat'],
                          lon_gridop = gridop['lon'], lat_gridop = gridop['lat'])


    @classmethod
    def from_frame(cls, df: pd.DataFrame) -> Self: 
        """Create a GridLonLat instance from a DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            A pandas DataFrame with a valid format (see Field documentation).

        Returns
        -------
        An instance of the grid subclass.
        """
        lon = np.sort(df.index.unique(level='lon').values)
        lat = np.sort(df.index.unique(level='lat').values)

        xmin = {}
        xmax = {}
        dx = {}
        gridop: dict['str', 'str' | None] = {}

        for axis_name, axis_values in zip(['lon', 'lat'], [lon, lat]):
            if len(axis_values) == 1:
                # Is it a reduced DataFrame?
                if isinstance(axis_values[0], str):
                    try:
                        get_gridop(grid_type, axis_values[0], axis_name)
                    except KeyError:
                        raise ValueError(f"Axis value '{axis_values[0]}' is not a spatial reduction.")
                    xmin[axis_name] = np.nan
                    xmax[axis_name] = np.nan
                    dx[axis_name] = np.nan
                    gridop[axis_name] = axis_values[0]
                else:
                    raise ValueError(f"Axis '{axis_name}' must be of size larger than 1.")
            else:
                xmin[axis_name] = axis_values.min()
                xmax[axis_name] = axis_values.max()
                dx[axis_name] = (axis_values[1:] - axis_values[:-1]).min()
                gridop[axis_name] = None

        return cls(lon_min=xmin['lon'], lon_max=xmax['lon'], dlon=dx['lon'],
                   lat_min=xmin['lat'], lat_max=xmax['lat'], dlat=dx['lat'],
                   lon_gridop=gridop['lon'], lat_gridop=gridop['lat'])


    def __add__(self, other):

        if not self.is_compatible(other):
            raise ValueError("Field grids are not compatible.")

        if self.is_reduced('lon') or self.is_reduced('lat'):
            raise NotImplementedError('Reduced grids cannot be combined (for now).')

        if self.is_reduced('lon'):
            lon_min = np.nan
            lon_max = np.nan
            dlon = np.nan
            lon_gridop = self.axis_gridop['lon']
        else:
            lon_min = min(self.xmin['lon'], other.xmin['lon'])
            lon_max = max(self.xmax['lon'], other.xmax['lon'])
            dlon = self.dx['lon']
            lon_gridop = None

        if self.is_reduced('lat'):
            lat_min = np.nan
            lat_max = np.nan
            dlat = np.nan
            lat_gridop = self.axis_gridop['lat']
        else:
            lat_min = min(self.xmin['lat'], other.xmin['lat'])
            lat_max = max(self.xmax['lat'], other.xmax['lat'])
            dlat = self.dx['lat']
            lat_gridop = None

        return GridLonLat(lon_min=lon_min, lon_max=lon_max, dlon=self.dx['lon'],
                          lat_min=lat_min, lat_max=lat_max, dlat=self.dx['lat'])


    def __repr__(self) -> str:
        repr_str = [super().__repr__()]
        if self.is_reduced('lon') or self.is_reduced('both'):
            repr_str.append(f"Longitude: reduced (gridop = {self.axis_gridop['lon']})")
        else:
            repr_str.append(f"Longitude: {self.lon_min} to {self.xmax['lon']} (step: {self.dx['lon']})")
        if self.is_reduced('lat') or self.is_reduced('both'):
            repr_str.append(f"Latitude: reduced (gridop = {self.axis_gridop['lat']})")
        else:
            repr_str.append(f"Latitude: {self.xmin['lat']} to {self.xmax['lat']} (step: {self.dx['lat']})")

        return '\n'.join(repr_str)
    

    def __str__(self) -> str:
        return self.__repr__()


    def is_compatible(self, other: Grid) -> bool:

        if type(other) is not GridLonLat:
            return False

        for axis in ['lon', 'lat']:
            if self.axis_gridop[axis] != other.axis_gridop[axis]:
                return False

            if self.is_reduced(axis):
                continue
            
            if self.dx[axis] != other.dx[axis]:
                return False

            x0_self = self.xmin[axis] - int((self.xmin[axis] - COORD_BOUNDARIES[axis])/self.dx[axis])*self.dx[axis]
            x0_other = other.xmin[axis] - int((other.xmin[axis] - COORD_BOUNDARIES[axis])/other.dx[axis])*other.dx[axis]

            if not np.isclose(x0_self, x0_other, rtol=1.e-7):
                return False

        return True


    @property
    def lon(self):
        return self.axes['lon']

    @property
    def dlon(self):
        return self.dx['lon']

    @property
    def lon_min(self):
        return self.xmin['lon']

    @property
    def lon_max(self):
        return self.xmax['lon']

    @property
    def lat(self):
        return self.axes['lat']

    @property
    def dlat(self):
        return self.dx['lat']

    @property
    def lat_min(self):
        return self.xmin['lat']

    @property
    def lat_max(self):
        return self.xmax['lat']


# ---------------
# GRID OPERATIONS
# ---------------

@register_gridop(grid_type)
def av_lon(df: pd.DataFrame, grid: GridLonLat) -> pd.DataFrame:
    """Average data along the longitude axis.

    Parameters
    ----------
    df : pd.DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : Grid
        A GridLonLat object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['lat', 'time']
    df_red = df.groupby(group_levels).mean()

    return df_red


@register_gridop(grid_type)
def av_lat(df: pd.DataFrame, grid: GridLonLat) -> pd.DataFrame:
    """Average data along the latitude axis.

    Parameters
    ----------
    df : pd.DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : GridLonLat
        A GridLonLat object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['lon', 'time']
    df_red = df.groupby(group_levels).mean()

    return df_red


@register_gridop(grid_type)
def av_both(df: pd.DataFrame, grid: GridLonLat) -> pd.DataFrame:
    """Average data across the whole domain.

    Each gridcell value is weighted by its corresponding area element:
        da = EARTH_RADIUS**2*dlon*dlat*cos(lat),

    where the angles are in radians.

    Parameters
    ----------
    df : pd.DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : GridLonLat
        A GridLonLat object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['time']
    weights = np.cos(df.index.get_level_values('lat').to_numpy()*DEG_TO_RAD)
    df_red = df.multiply(weights, axis='index')
    df_red['w'] = weights
    df_red = df_red.groupby(group_levels).sum()
    df_red = df_red.div(df_red['w'], axis='index')
    df_red.drop('w', axis=1, inplace=True)

    return df_red


@register_gridop(grid_type)
def sum_lon(df: pd.DataFrame, grid: GridLonLat) -> pd.DataFrame:
    """Aggregate data along the longitude axis.

    Parameters
    ----------
    df : pd.DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : GridLonLat
        A GridLonLat object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    if grid.is_reduced('lat'):
        raise ValueError("Cannot calculate 'sum' along 'lon' axis on a field with a reduced 'lat' axis.")
    group_levels = ['lat', 'time']
    weights = EARTH_RADIUS*grid.dlon*DEG_TO_RAD*np.cos(df.index.get_level_values('lat').to_numpy()*DEG_TO_RAD)
    df_red = df.multiply(weights, axis='index').groupby(group_levels).sum()

    return df_red


@register_gridop(grid_type)
def sum_lat(df: pd.DataFrame, grid: GridLonLat) -> pd.DataFrame:
    """Aggregate data along the latitude axis.

    Parameters
    ----------
    df : pandas DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : GridLonLat
        A GridLonLat object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['lon', 'time']
    df_red = (EARTH_RADIUS*grid.dlat*DEG_TO_RAD*df).groupby(group_levels).sum()

    return df_red


@register_gridop(grid_type)
def sum_both(df: pd.DataFrame, grid: GridLonLat) -> pd.DataFrame:
    """Aggregate data across the whole domain.

    Each gridcell value is weighted by its corresponding area element:
        da = EARTH_RADIUS**2*dlon*dlat*cos(lat),
        
    where the angles are in radians.

    Parameters
    ----------
    df: pandas DataFrame
        The pandas DataFrame whose data is to be averaged.
    grid : GridLonLat
        A GridLonLat object.

    Returns
    -------
    A reduced pandas DataFrame
    """
    group_levels = ['time']
    weights = EARTH_RADIUS**2*DEG_TO_RAD**2*grid.dlat*grid.dlon*np.cos(df.index.get_level_values('lat').to_numpy()*DEG_TO_RAD)
    df_red = df.multiply(weights, axis='index')
    df_red = df_red.groupby(group_levels).sum()

    return df_red




from dataclasses import dataclass
import json
import os
from canopy.core.field import Field

@dataclass
class TestDataInfo:
    description: str = ""
    file_format: str = ""
    grid_type: str = ""
    source: str = ""

_test_data_info = {}


def _load_test_data_registry():

    with open(f'{os.path.dirname(__file__)}/registry.json') as f:
        test_data = json.load(f)
    for fname, data in test_data.items():
        td = TestDataInfo(description=data['description'],
                          file_format=data['file_format'],
                          grid_type=data['grid_type'],
                          source=data['source']
                          )
        _test_data_info[fname] = td


def get_test_data(fname: str | None = None) -> Field | None:
    '''Load a test data field

    Parameters
    ----------
    fname : str | None
        The name of the data file to load. If None, a list of available test data files is printed, and the function returns None

    Returns
    -------
        Either a Field object derived from the data file or None.
    '''

    if fname is None:
        msg = [
                "Available test data:",
                "--------------------",
                ]

        for fname, info in _test_data_info.items():
            msg.append(f"{fname}: {info.description}")

        print("\n".join(msg))
        return None

    path = f'{os.path.dirname(__file__)}'
    tdi = _test_data_info[fname]
    field = Field.from_file(f'{path}/{fname}', grid_type=tdi.grid_type, source=tdi.source)

    return field


_load_test_data_registry()





import copy
import pandas as pd
import xarray as xr
import numpy as np
from datetime import datetime
from typing import Optional, Union, Sequence, Callable, Literal, SupportsFloat, cast
from typing_extensions import Self
from canopy.readers.registry import get_reader, get_format_description
from canopy.grid import get_grid, get_grid_type
from canopy.grid.grid_abc import Grid
from canopy.grid.grid_empty import GridEmpty
from canopy.core.redspec import RedSpec
from canopy.source_data import get_source_data
import canopy.core.frameops as frameops
import janitor


class FieldData:
    """Validator class for the data parameter passed to Field"""

    private_name = '_data'

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, data):
        # Protect data from resetting
        if self.private_name in vars(obj):
            raise AttributeError("'Field.data' cannot be set.")
        self.validate(data)
        setattr(obj, self.private_name, data)

    def validate(self, data):
        # The data must be an instance of pandas DataFrame
        if not isinstance(data, pd.DataFrame):
            raise ValueError("'data' must be a Pandas DataFrame object.")
        # The index must be a MultiIndex
        if not isinstance(data.index, pd.MultiIndex):
            raise ValueError("'data' must have a MultiIndex.")
        # MultiIndex must have at least 3 levels
        if len(data.index.names) < 3:
            raise ValueError("data's MultiIndex must have at least 3 levels.")
        # The first and second levels of the MultiIndex must be of numeric type
        # TODO: This doesn't work for reduced lon/lat. Think about this problem
        #for ilevel in range(2):
        #    if not pd.api.types.is_numeric_dtype(data.index.get_level_values(ilevel).dtype):
        #        raise ValueError(f"Level {ilevel} of data's MultiIndex must be of numeric dtype.")
        # The third level's name must be "time"
        if data.index.names[2] != 'time':
            raise ValueError("The name of the third level of data's MultiIndex must be 'time'.")
        # 'time' level must be of type PeriodIndex
        if not isinstance(data.index.get_level_values(2), pd.PeriodIndex):
            raise ValueError("Level 'time' of data's Index must be of PeriodIndex type.")


class Field:
    """Container for data derived from model output or observations

    This object contains model output or observation data and associated
    grid information. It allows for basic data manipulation, such as 
    time- and spatial reductions and slicing. The Field object is canopy's
    elemental interface between data and user.
    """

    data = FieldData()

    def __init__(self, data: pd.DataFrame, grid: Grid, modified: bool = False) -> None:
        """
        Parameters
        ----------
        data : pd.DataFrame
            Pandas DataFrame containing the Field's data (see specification below).
        grid : Grid
            A canopy Grid object (see Grid object documentation).
        modified : bool
            Whether the field being created has been modified (e.g. reduced).
        """

        self.data = data
        self._grid: Grid = grid
        self._modified: bool = modified
        self._metadata: dict[str, str | float | int] = {'name': '[no name]',
                                                        'units': '[no units]',
                                                        'description': '[no description]',
                                                        }
        self._history: list[str] = []
        self._timeop: str | None = None
        time_index = cast(pd.PeriodIndex, data.index.get_level_values('time'))
        self._time_freq: str = str(time_index.freqstr)

        # Order MultiIndex, necessary for slicing
        if self.data.index._lexsort_depth < 3:
            self._data = self._data.sort_index(level=0)


    def __getitem__(self, layers):
        """
        Invoke `self.sel_layers` with indexing notation
        """
        return self.sel_layers(layers)

    
    def __setitem__(self, layers: str | Sequence[str], value: Union[SupportsFloat, np.ndarray, pd.DataFrame, 'Field']) -> None:

        if isinstance(layers, str):
            layers = [layers]

        if isinstance(value, SupportsFloat):
            self.data[layers] = float(value)
            what = f'the scalar {value}'
        elif isinstance(value, Field):
            frameops.check_indices_match(self.data, value.data)
            self.data[layers] = value.data.copy()
            what = f'the following layers from another Field: {value.layers}'
        else:
            raise ValueError("Can only assign another Field or a scalar.")
        
        self._modified = True
        log_message = f'Layers {list(layers)} were assigned {what}.'
        self.log(log_message)


    @property
    def grid(self) -> Grid:
        """Grid: The grid associated with the data"""
        return self._grid


    @property
    def modified(self) -> bool:
        """bool: True if the Field has been modified after loading."""
        return self._modified


    @property
    def metadata(self) -> dict[str, str | float | int]:
        """dict: Field's metadata (units, etc...)"""
        return self._metadata


    @property
    def history(self) -> list[str]:
        """list: Keeps the history of modifications of the Field"""
        return self._history

    
    @property
    def timeop(self) -> str | None:
        """str: returns the reduction operation applied to the time dimension (if any)."""
        return self._timeop


    @property
    def time_freq(self) -> str:
        """str: returns the sampling frequency of the time axis."""
        return self._time_freq


    @property
    def layers(self) -> list[str]:
        """str: returns a list with the Field's layer names."""
        return list(self.data.columns)


    @classmethod
    def from_file(cls,
                  path: str,
                  file_format: str | None = None,
                  grid_type: str | None = None, 
                  grid: Grid | None = None,
                  source: str | None = None,
                  **kwargs) -> Self:
        """Construct a Field object from an LPJ-GUESS output file.

        Parameters
        ----------
        path : str
            Path to output file.
        file_format : str
            One of the registered formats (see file_readers.py). For LPJ-GUESS standard output:
                - 'lpjg_annual' (annual output)
                - 'lpjg_monthly' (monthly output)
        grid : Grid
            A Grid object describing the grid associated to the data. If no Grid is passed, it is inferred from the dataframe.
        grid_type : str
            The type of grid associated to the data. Accepted values are ['sites', 'lonlat']
        source : str
            The source to retrieve the file's metadata. The format for this argument is 'source:field'. For example, to read
            an LPJ-GUESS file and add metadata corresponding to Annual GPP:

            agpp = Field.from_file('/path/to/file/file_name.out', file_format='lpjg_annual', source='lpjguess:agpp')
        kwargs
            The kwargs are passed to the file reader.

        Returns
        -------
        A Field object.

        Notes
        -----
        - When a Grid object is provided (via the grid argument), the Field is always created
          with this grid, regardless of the value of 'grid_type', which is overriden. It is the
          responsibility of the caller to provide a consistent grid object.
        - Otherwise, a Grid object is constructed. If grid_type == 'sites', a default Grid (no grid information) is created.
          Otherwise, a grid is inferred from the data.
        """

        if grid_type is None:
            grid_type = "lonlat"

        if file_format is None:
            file_format = "lpjg_annual"

        # TODO: I shouldn't need to pass grid_type other than in the kwargs to a reader
        # A reader should be Callable[[str, **kwargs], pd.DataFrame]
        df = get_reader(file_format)(path, grid_type, **kwargs)
        format_desc = get_format_description(file_format)

        # If a grid is provided, it should be used:
        if grid is not None:
            pass
        # If it is not provided, it needs to be constructed
        else:
            grid = get_grid(grid_type).from_frame(df)

        field = cls(df, grid) 
        field.log(f"Data read from {path}")
        field.add_md('file format', format_desc)
        field.add_md('original file', path)
        if source is not None:
            src, fld = source.split(':')
            source_data = get_source_data(src)
            field.set_md('name', source_data['fields'][fld]['name'])
            field.set_md('description', source_data['fields'][fld]['description'])
            field.set_md('units', source_data['fields'][fld]['units'])

        return field


    def add_md(self, key: str, value: str | float | int) -> None:
        """Add an entry to the metadata dictionary.

        Parameters
        ----------
        key : str
            The key under which the value is stored.
        value : str | float | int
            The value of the metadata entry.

        Example
        -------
        # Load field
        anpp = Field.from_file("/path/to/file/anpp.out")
        # Add the GCM used to force the simulation to the metadata
        field.add_md('gcm', 'MPIESM2.1')

        Notes
        -----
        If the key already exists in the metadata dictionary, a KeyError will be raised. In
        order to overwrite an entry, use the set_md method.
        """
        if key in self._metadata:
            raise KeyError(f"Key {key} already exists (use set_md method to replace it)")
        else:
            self._metadata[key] = value


    def set_md(self, key: str, value: str | float | int) -> None:
        """Add or replace an entry in the metadata dictionary.

        Parameters
        ----------
        key : str
            The key under which the value is stored.
        value : str | float | int
            The value of the metadata entry.

        Example
        -------
        # Load field
        anpp = Field.from_file("/path/to/file/anpp.out")
        # Fails because 'name' is set in the constructor
        field.add_md('name', 'NPP') # KeyError
        field.set_md('name', 'NPP') # Okay
        """
        self._metadata[key] = value


    def copy_md(self, field: 'Field') -> None:
        """Replace the metadata with a copy of the passed Field's.

        Parameters
        ----------
        field : Field
            The field whose metadata is to be copied.
        """
        self._metadata = copy.deepcopy(field.metadata)


    def log(self, entries: str | list[str]) -> None:
        """File one or more entries in the Field's history log.

        The function files the passed entry in the history log and
        adds a timestamp.

        Parameters
        ----------
        entries : str | list[str]
            The entry or list of entries to log. If a list is provided,
            the same timestamp will be attached to all entries.
        """
        if isinstance(entries, str):
            entries = [entries]
        dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        for entry in entries:
            self._history.append(f"[{len(self.history) + 1}] {dt}: {entry}")


    def copy_history(self, field: 'Field') -> None:
        """Replace history with a copy of the history of the passed Field.

        Parameters
        ----------
        field : Field
            The field whose history is to be copied.
        """
        self._history = copy.deepcopy(field.history)


    def sel_slice(self, lon_slice: Optional[tuple[float,float]] = None,
                        lat_slice: Optional[tuple[float,float]] = None,
                        time_slice: Optional[tuple[int,int]] = None,
                        inplace: bool = False) -> Union[None,'Field']:
        """Select a time or spatial slice from the Field.

        Parameters
        ----------
        lon_slice : Optional[tuple[float,float]] = None
            Tuple or list of floats specifying a slice in longitude.
        lat_slice : Optional[tuple[float,float]] = None
            Tuple or list of floats specifying a slice in latitude.
        time_slice : Optional[tuple[int,int]] = None
            Tuple or list of ints specifying a year range.
        inplace : bool = False
            If True, the slicing is performed in place.

        Returns
        -------
        If inplace is True, the slicing is performed in place and the method returns None.
        Otherwise the method returns a Field object containing the sliced data.
        """

        #TODO: The time slice should be based on datetime objects
        df, grid, log_message = frameops.sel_slice(self.data, self.grid, lon_slice, lat_slice, time_slice)

        if inplace:
            if df.index._lexsort_depth < 3:
                df = df.sort_index(level=0)
            self._data = df
            self._grid = grid
            self._modified = True
            self.log(log_message)
            return None
        else:
            field = Field(df, grid, modified = True)
            field.copy_history(self)
            field.copy_md(self)
            field.log(log_message)
            return field


    def sel_layers(self, layers: str | Sequence[str] | pd.Index, inplace: bool = False) -> Union[None,'Field']:
        """Select one or more layers from the Field.

        Parameters
        ----------
        layers: str | Sequence[str]
            Layer name or list of layer names.
        inplace : bool = False
            If True, the selection is performed in place.

        Returns
        -------
        If inplace is True, the selection is performed in place and the method returns None.
        Otherwise the method returns a Field object containing the selected data.
        """

        if isinstance(layers, str):
            layers = [layers]

        log_message=f"Selected layers {list(layers)}."
        if inplace:
            self._data = self._data[layers]
            self._modified = True
            self.log(log_message)
            return None
        else:
            df = self._data[layers].copy()
            grid = copy.deepcopy(self._grid)
            field = Field(df, grid, modified=True)
            field.copy_history(self)
            field.copy_md(self)
            field.log(log_message)
            return field


    def sel_overlap(self, other: 'Field', exact: bool = True, atol: float = 0.1, use_other_coords: bool = False, inplace: bool = False) -> Union[None,'Field']:

        if exact and not self.grid.is_compatible(other.grid):
            raise ValueError("Exact overlap selection requires that the fields have compatible grids.")

        for ax_self, ax_other in zip(self.data.index.names[:2], other.data.index.names[:2]):
            if ax_self != ax_other:
                raise ValueError("Overlap selection requires that spatial axes have the same names for both fields.")

        if not (self.time_freq == other.time_freq):
            raise ValueError("Overlap selection requires that the time series of the fields have the same frequency.")

        if exact:
            df = frameops.select_overlap_exact(self.data, other.data)
            how = "exactly"
        else:
            df = frameops.select_overlap_with_spatial_tolerance(self.data, other.data, atol, use_other_coords)
            how = f"with tolerance {atol} in the spatial coordinates"

        if df.empty:
            log_message = "sel_overlap: Field was sliced to empty"
            grid = GridEmpty()
        else:
            log_message = f"sel_overlap: sliced with overlapping rows from another field (description: {other.metadata['description']}). The index intersection was calculated {how}."
            if use_other_coords:
                grid = get_grid(other.grid.grid_type).from_frame(df)
            else:
                grid = get_grid(self.grid.grid_type).from_frame(df)

        if inplace:
            if df.index._lexsort_depth < 3:
                df = df.sort_index(level=0)
            self._data = df
            self._modified = True
            self._grid = grid
            self.log(log_message)
            return None
        else:
            field = Field(df, grid, modified=True)
            field.copy_history(self)
            field.copy_md(self)
            field.log(log_message)
            return field


    def drop_layers(self, layers: str | Sequence[str] | pd.Index, inplace: bool = False) -> Union[None,'Field']:
        """Drop one or more layers from the Field.

        Parameters
        ----------
        layers : str | list
            Layer name or list of layer names.
        inplace : bool = False
            If True, the layers are dropped from the current object.

        Returns
        -------
        If inplace is True, the layers are dropped from the current Field and the method returns None.
        Otherwise the method returns a Field object with the selected data.
        """

        if isinstance(layers, str):
            layers = [layers]

        log_message=f"Dropped layers {list(layers)}."
        if inplace:
            self._data.drop(layers, axis=1, inplace=True)
            self._modified = True
            self.log(log_message)
            return None
        else:
            df = self._data.drop(layers, axis=1)
            grid = copy.deepcopy(self._grid)
            field = Field(df, grid, modified=True)
            field.copy_history(self)
            field.log(log_message)
            return field


    def red_layers(self, redop: str, layers: Optional[Sequence[str] | pd.Index] = None,
                   name: Optional[str] = None, drop: bool = False, inplace: bool = False) -> Union[None,'Field']:
        """Perform a reduction operation on the Field's layers.

        Parameters
        ----------
        redop : str
            The reduction operation. One of 'sum', 'av', 'maxLay', '/'.
        layers : None or a list of strings.
            List of names of layers to be reduced. If None, all layers are reduced.
        name : None or str.
            Name of the new layer to store the reduction. If None, the redop argument is used.
        drop : bool
            If True, the reduced layers are dropped from the data.
        inplace : bool
            If True, the layers are reduced in the current object.

        Returns
        -------
        If inplace is True, the layers are reduced in the current Field and the method returns None.
        Otherwise the method returns a Field object with the reduced data.
        """
        if layers is None:
            layers = self._data.columns

        # Copy avoids warning (error)
        df_red = self._data.copy()

        if name is None:
            name = f"{redop}"

        if redop == 'sum':
            df_red[name] = df_red[layers].sum(axis=1)
        if redop == 'av':
            df_red[name] = df_red[layers].mean(axis=1)
        if redop == 'maxLay':
            df_red[name] = df_red[layers].idxmax(axis=1)
        if redop == '/':
            if not len(layers) == 2:
                raise ValueError(f"Operation '{redop}' requires exactly two operands")
            df_red[name] = df_red[layers[0]] / df_red[layers[1]]
        if drop:
            layers_to_drop = [l for l in layers if l != name]
            df_red.drop(layers_to_drop, axis=1, inplace=True)

        log_message=f"Layer reduction operation '{redop}' applied to layers {list(layers)}, stored as layer '{name}'.{' Original layers were dropped.' if drop else ''}"
        if inplace:
            self._data = df_red
            self._modified = True
            self.log(log_message)
            return None
        else:
            grid = copy.deepcopy(self._grid)
            field = Field(df_red, grid, modified=True)
            field.copy_history(self)
            field.log(log_message)
            return field


    def filter(self, query: str, fill_nan: bool = False, inplace: bool = False) -> Union[None, 'Field']:
        '''Filter rows based on boolean query

        Parameters
        ----------
        query: str
            The string describing the boolean query in terms of the index or layers
        fill_nan: bool
            If False, rows where the query is False are removed (default behaviour).
            If True, the resulting field's data has NaNs where the query is false.
        inplace: bool
            Whether to perform the operation in place

        Returns
        -------
        If inplace is True, the filtering is performed on the current Field and the method returns None.
        Otherwise the method returns a Field object with the reduced data.

        Example
        --------
        # Load field
        aaet = Field.from_file("/path/to/file/aaet.out")
        print(aaet.layers) # [..., 'Total']
        # Filter rows for which layer 'Total' is greater than 100
        aaet1 = aaet.filter('Total > 100')
        # Filter rows for which layer 'Total' is greater than 100 and layer 'C3G' is lower than 10 (inplace)
        aaet.filter('Total > 100 and C3G < 10', inplace=True)

        Notes
        -----
        See pandas documentation for DataFrame.query() for more details on the query string.
        '''
        if fill_nan:
            index = self.data.query(f'not ({query})').index
            if inplace:
                self.data.loc[index,:] = np.nan
                df = self.data
                self._modified = True
            else:
                df = self.data.copy()
                df.loc[index,:] = np.nan
        else:
            if inplace:
                self.data.query(query, inplace=True)
                df = self.data
                self._modified = True
            else:
                df = self.data.query(query)

        log_message = f"Filter data: query = '{query}'{', fill with NaNs.' if fill_nan else '.'}"

        if df.empty and not fill_nan:
            grid: Grid = GridEmpty()
            log_message += ' Filter operation yielded empty field.'
        else:
            grid = get_grid(self.grid.grid_type).from_frame(df)
        
        if inplace:
            self._data = df
            self._grid = grid
            self._modified = True
            self.log(log_message)
            return None
        else:
            field = Field(df, grid, modified=True)
            field.copy_history(self)
            field.copy_md(self)
            field.log(log_message)
            return field
        

    def red_time(self, timeop: str, freq: str | None = None, inplace: bool = False) -> Union[None,'Field']:
        """Perform a reduction operation on the time axis.

        Parameters
        ----------
        timeop : str
            The time reduction operation: one of 'av', 'sum'.
        freq : str | None = None
            A string specifying the frequency of the reduction. This is formed by an integer number and
            one of 'M', 'Y'. For example, to perform an average every five years, specify timeop='av' and
            freq='5Y'. If freq=None the whole time series is reduced.
        inplace : bool = False
            If True, the reduction is performed on the current Field.
            
        Returns
        -------
        If inplace is True, the reduction is performed on the current Field and the method returns None.
        Otherwise the method returns a Field object with the reduced data.
        """
        df_red, log_message = frameops.red_time(self._data, timeop, freq)

        if inplace:
            self._data = df_red
            self._timeop = timeop
            self._modified = True
            self.log(log_message)
            return None
        else:
            field_red = Field(df_red, copy.deepcopy(self._grid), modified=True)
            field_red._timeop = timeop
            field_red.copy_history(self)
            field_red.log(log_message)
            return field_red


    def red_space(self, gridop: str, axis: Optional[str] = None, inplace: bool = False) -> Union[None,'Field']:
        """Perform a reduction operation on the spatial axes.

        Parameters
        ----------
        gridop : str
            The spatial reduction operation: one of 'av', 'sum'.
        axis : str | None = None
            One of 'lon', 'lat', or 'both'. If not specified, the reduction is performed on both axes.
        inplace : bool = False
            If True, the reduction is performed on the current Field.
            
        Returns
        -------
        If inplace is True, the reduction is performed on the current Field and the method returns None.
        Otherwise the method returns a Field object with the reduced data.
        """

        if axis is None:
            axis = 'both'

        df_red, grid_red, log_message = frameops.red_space(self.data, self.grid, gridop, axis)

        if inplace:
            self._data = df_red
            self._grid = grid_red
            self._modified = True
            self.log(log_message)
            return None
        else:
            field_red = Field(df_red, grid_red, modified=True)
            field_red.copy_history(self)
            field_red.log(log_message)
            return field_red


    def reduce(self, redspec: RedSpec, inplace: bool = False) -> Union[None,'Field']:
        """Perform the selection/slicing/reduction operations specified in te passed RedSpec object.

        Parameters
        ----------
        redspec : RedSpec
            A RedSpec object specifying how to slice and/or reduce the data.
        inplace : bool = False
            If True, the reduction is performed on the current Field.
            
        Returns
        -------
        If inplace is True, the reduction is performed on the current Field and the method returns None.
        Otherwise the method returns a Field object with the reduced data.
        """
        df_red, grid_red, log_message = frameops.apply_reduction(self._data, self._grid, redspec)

        time_index = cast(pd.PeriodIndex, df_red.index.get_level_values('time'))
        if inplace:
            self._data = df_red
            self._grid = grid_red
            self._modified = True
            self.log(log_message)
            if redspec.timeop is not None:
                self._timeop = redspec.timeop
                setattr(self, '_time_freq', str(time_index.freqstr))
            return None
        else:
            field_red = Field(df_red, grid_red, modified=True)
            field_red.copy_history(self)
            field_red.log(log_message)
            return field_red


    def convert_units(self, factor: SupportsFloat, units: str, inplace: bool = False) -> None:
        """Convert the field's units by a multiplicative factor

        Parameters
        ----------
        factor : SupportsFloat
            Scalar by which all values in the field's data are multiplied
        units : str
            The new units to be set in the metadata 'units' entry.
        inplace : bool = False
            If True, the reduction is performed on the current Field.

        Notes
        -----
        This function does not perform any checks on whether the passed factor or
        units string make sense! It just trusts that the user knows what they are doing.
        """
        try:
            df = self._data*float(factor)
        except ValueError:
            raise TypeError("'factor' must be of numeric type, or convertible to 'float'.")

        log_message = f'Units changed to {units} (factor: {factor}).'

        if inplace:
            self._data = df
            self.set_md('units', units)
            self._modified = True
            self.log(log_message)
        else:
            grid = copy.deepcopy(self.grid)
            field = Field(df, grid, modified=True)
            field.copy_history(self)
            field.set_md('units', units)
            field.log(log_message)
            return field


    def rename_layers(self, new_names: dict[str, str]) -> None:
        """Rename the field's layers

        Parameters
        ----------
        new_names : dict
            A dictionary mapping existing layer names to their new names.
        """
        self.data.rename(columns=new_names, inplace=True)
        self.log(f"Layers renamed: {new_names}")


    def apply(self, op: str | Callable, operand: SupportsFloat | list[str] | None = None, layers: str | list[str] | None = None, how: Literal['left', 'right'] = 'left', inplace: bool = False) -> Union[None,'Field']:
        """Apply an operation/function to selected layers

        Parameters
        ----------
        op : str | Callable
            Operation to apply to layers. If a callable is passed, it must be numpy-vectorizable.
        operand : str | list[str] | None
            Operand to combine with layers through 'op'. Operand can be a constant number or the name of a layer. In the latter
            case, the operation will be performed element-wise. If 'op' is a Callable, this parameter is ignored.
        layers : str | list[str] | None
            A list of the names of the layers to apply the operation to. If None, the operation is applied to all layers.
        how : str = 'left'
            Position of the layers in the operation. This argument is relevant for non-commutative operations. For example, if
            how == 'left', a '-' operation will be layers - operand. If how = 'right', the operation will be operand - layers.
        inplace : bool
            Whether to perform the operation in place.

        Returns
        -------
        A field with the modified layers, or None if the operation is performed in place.
        """

        commutative_ops = {
            '+':   frameops.apply_sum,
            'sum': frameops.apply_sum,
            '*':   frameops.apply_mul,
            'mul': frameops.apply_mul,
        }

        non_commutative_ops = {
            '-':   frameops.apply_sub,
            'sub': frameops.apply_sub,
            '/':   frameops.apply_div,
            'div': frameops.apply_div,
        }

        if layers is None:
            layers = self.layers
        elif isinstance(layers, str):
            layers = [layers]

        if isinstance(operand, str):
            operand = [operand]

        if (op in commutative_ops or op in non_commutative_ops) and operand is None:
            raise ValueError(f"Operation {op} requires an operand")

        if inplace:
            df = self.data
        else:
            df = self.data.copy()

        if callable(op):
            frameops.apply_function(df, layers, op)
        elif op in commutative_ops:
            # Ignoring typing; at this point it's alredy known that operand is not None, but mypy's narrowing doesn't catch it
            commutative_ops[op](df, operand, layers) # type: ignore
        elif op in non_commutative_ops:
            # Ignoring typing; at this point it's alredy known that operand is not None, but mypy's narrowing doesn't catch it
            non_commutative_ops[op](df, operand, layers, how) # type: ignore
        else:
            raise ValueError(f"Operation must be a callable or one of {list(commutative_ops.keys()) + list(non_commutative_ops.keys())}.")

        log_message = f"Applied operation '{op}' to layers {layers}."

        if not inplace:
            grid = copy.deepcopy(self.grid)
            field = Field(df, grid, modified=True)
            field.copy_history(self)
            field.log(log_message)
            return field
        else:
            self._modified = True
            self.log(log_message)
            return None


    def __str__(self) -> str:

        history_str = '\n'.join(self.history)

        if self.data.empty:
            repr_str = [
                    f"Field is empty!",
                    "",
                    f"History",
                    f"-------",
                    history_str,
                    ]

            return '\n'.join(repr_str)

        metadata_str = '\n'.join([f"{k}: {v}" for k, v in self.metadata.items()])
        grid_type = get_grid_type(self._grid)

        time_index = self.data.index.get_level_values('time')
        time_start = time_index.min().start_time
        time_end = time_index.max().end_time
        time_str = [f"Span: {time_start} - {time_end}",
                    f"Frequency: {self.time_freq}",
                    ]
        if self._timeop is not None:
            time_str.append(f"Reduction: '{self._timeop}'")

        repr_str = [
                    f"Data",
                    f"----",
                    metadata_str,
                    "",
                    f"Grid",
                    f"----",
                    f"{self._grid}",
                    "\n"
                    f"Time series",
                    f"-----------",
                    '\n'.join(time_str),
                    "\n"
                    f"History",
                    f"-------",
                    history_str,
                    ]
        return "\n".join(repr_str)


    def __repr__(self) -> str:
        return self.__str__()


    def to_xarray(self):
        raise NotImplementedError
        #return self._data.to_xarray().sortby('lon').sortby('lat').sortby('time')


    def to_netcdf(self, fname: Optional[str] = None):
        raise NotImplementedError
        # TODO: This doesn't work w/ PeriodIndex!
        #data_xr = to_xarray(self._data)
        #data_xr.to_netcdf(fname)


    def to_csv(self, fname: str):
        raise NotImplementedError
        #self._data.to_csv(fname)

